package net.awired.ajsl.core.collect;

import java.io.Serializable;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import com.google.common.base.Preconditions;

/**
 * A list auto populated if get by index and remove null index on iteration.
 */
public class PopulatingShrinkList<E> implements List<E>, Serializable {

    private static final long serialVersionUID = 1L;

    /**
     * The {@link List} that all operations are eventually delegated to.
     */
    private final List<E> backingList;

    /**
     * The {@link ElementFactory} to use to create new {@link List} elements
     * on demand.
     */
    private final ElementFactory<E> elementFactory;

    /**
     * Decorates list with shrinkable lazy list.
     */
    public static <E> List<E> decorate(Class<E> elementClass) {
        return new PopulatingShrinkList<E>(elementClass);
    }

    /**
     * Creates a new <code>AutoPopulatingList</code> that is backed by a standard {@link ArrayList} and adds new
     * instances of the supplied {@link Class element Class} to the backing {@link List} on demand.
     */
    public PopulatingShrinkList(Class<? extends E> elementClass) {
        this(new ArrayList<E>(), elementClass);
    }

    /**
     * Creates a new <code>AutoPopulatingList</code> that is backed by the supplied {@link List} and adds new instances
     * of the supplied {@link Class element Class} to the backing {@link List} on demand.
     */
    public PopulatingShrinkList(List<E> backingList, Class<? extends E> elementClass) {
        this(backingList, new ReflectiveElementFactory<E>(elementClass));
    }

    /**
     * Creates a new <code>AutoPopulatingList</code> that is backed by a standard {@link ArrayList} and creates new
     * elements on demand using the supplied {@link ElementFactory}.
     */
    public PopulatingShrinkList(ElementFactory<E> elementFactory) {
        this(new ArrayList<E>(), elementFactory);
    }

    /**
     * Creates a new <code>AutoPopulatingList</code> that is backed by the supplied {@link List} and creates new
     * elements on demand using the supplied {@link ElementFactory}.
     */
    public PopulatingShrinkList(List<E> backingList, ElementFactory<E> elementFactory) {
        Preconditions.checkNotNull(backingList, "Backing List must not be null");
        Preconditions.checkNotNull(elementFactory, "Element factory must not be null");
        this.backingList = backingList;
        this.elementFactory = elementFactory;
    }

    @Override
    public void add(int index, E element) {
        this.backingList.add(index, element);
    }

    @Override
    public boolean add(E o) {
        return this.backingList.add(o);
    }

    @Override
    public boolean addAll(Collection<? extends E> c) {
        return this.backingList.addAll(c);
    }

    @Override
    public boolean addAll(int index, Collection<? extends E> c) {
        return this.backingList.addAll(index, c);
    }

    @Override
    public void clear() {
        this.backingList.clear();
    }

    @Override
    public boolean contains(Object o) {
        return this.backingList.contains(o);
    }

    @Override
    public boolean containsAll(Collection c) {
        return this.backingList.containsAll(c);
    }

    /**
     * Get the element at the supplied index, creating it if there is
     * no element at that index.
     */
    @Override
    public E get(int index) {
        int backingListSize = this.backingList.size();
        E element = null;
        if (index < backingListSize) {
            element = this.backingList.get(index);
            if (element == null) {
                element = this.elementFactory.createElement(index);
                this.backingList.set(index, element);
            }
        } else {
            for (int x = backingListSize; x < index; x++) {
                this.backingList.add(null);
            }
            element = this.elementFactory.createElement(index);
            this.backingList.add(element);
        }
        return element;
    }

    public void shrink() {
        for (Iterator<E> i = backingList.iterator(); i.hasNext();) {
            if (i.next() == null) {
                i.remove();
            }
        }
    }

    @Override
    public int indexOf(Object o) {
        return this.backingList.indexOf(o);
    }

    @Override
    public boolean isEmpty() {
        return this.backingList.isEmpty();
    }

    @Override
    public Iterator<E> iterator() {
        shrink();
        return this.backingList.iterator();
    }

    @Override
    public int lastIndexOf(Object o) {
        return this.backingList.lastIndexOf(o);
    }

    @Override
    public ListIterator<E> listIterator() {
        return this.backingList.listIterator();
    }

    @Override
    public ListIterator<E> listIterator(int index) {
        return this.backingList.listIterator(index);
    }

    @Override
    public E remove(int index) {
        return this.backingList.remove(index);
    }

    @Override
    public boolean remove(Object o) {
        return this.backingList.remove(o);
    }

    @Override
    public boolean removeAll(Collection<?> c) {
        return this.backingList.removeAll(c);
    }

    @Override
    public boolean retainAll(Collection<?> c) {
        return this.backingList.retainAll(c);
    }

    @Override
    public E set(int index, E element) {
        return this.backingList.set(index, element);
    }

    @Override
    public int size() {
        return this.backingList.size();
    }

    @Override
    public List<E> subList(int fromIndex, int toIndex) {
        return this.backingList.subList(fromIndex, toIndex);
    }

    @Override
    public Object[] toArray() {
        return this.backingList.toArray();
    }

    @Override
    public <T> T[] toArray(T[] a) {
        return this.backingList.toArray(a);
    }

    @Override
    public boolean equals(Object other) {
        return this.backingList.equals(other);
    }

    @Override
    public int hashCode() {
        return this.backingList.hashCode();
    }

    /**
     * Factory interface for creating elements for an index-based access
     * data structure such as a {@link java.util.List}.
     */
    public interface ElementFactory<E> {

        /**
         * Create the element for the supplied index.
         * 
         * @return the element object
         * @throws ElementInstantiationException
         *             if the instantiation process failed
         *             (any exception thrown by a target constructor should be propagated as-is)
         */
        E createElement(int index) throws ElementInstantiationException;
    }

    /**
     * Exception to be thrown from ElementFactory.
     */
    public static class ElementInstantiationException extends RuntimeException {

        public ElementInstantiationException(String msg) {
            super(msg);
        }
    }

    /**
     * Reflective implementation of the ElementFactory interface,
     * using <code>Class.newInstance()</code> on a given element class.
     * 
     * @see java.lang.Class#newInstance()
     */
    private static class ReflectiveElementFactory<E> implements ElementFactory<E>, Serializable {

        private final Class<? extends E> elementClass;

        public ReflectiveElementFactory(Class<? extends E> elementClass) {
            Preconditions.checkNotNull(elementClass, "Element clas must not be null");
            Preconditions.checkState(!elementClass.isInterface(), "Element class must not be an interface type");
            Preconditions.checkState(!Modifier.isAbstract(elementClass.getModifiers()),
                    "Element class cannot be an abstract class");
            this.elementClass = elementClass;
        }

        @Override
        public E createElement(int index) {
            try {
                return this.elementClass.newInstance();
            } catch (InstantiationException ex) {
                throw new ElementInstantiationException("Unable to instantiate element class ["
                        + this.elementClass.getName() + "]. Root cause is " + ex);
            } catch (IllegalAccessException ex) {
                throw new ElementInstantiationException("Cannot access element class [" + this.elementClass.getName()
                        + "]. Root cause is " + ex);
            }
        }
    }

}
